/*
 * FreeSWITCH Modular Media Switching Software Library / Soft-Switch Application
 * Copyright (C) 2005-2014, Anthony Minessale II <anthm@freeswitch.org>
 *
 * Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is FreeSWITCH Modular Media Switching Software Library / Soft-Switch Application
 *
 * The Initial Developer of the Original Code is
 * Piotr Gregor <piotr@freeswitch.com> <piotr@signalwire.com>
 *
 * Portions created by the Initial Developer are Copyright (C)
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Piotr Gregor <piotr@freeswitch.org> <piotr@signalwire.com>
 *
 * mod_red.c -- The redundancy encoder/decoder (RFC 2198, RFC 5109)
 *
 */

#ifndef SWITCH_RED_H
#define SWITCH_RED_H

#include <switch_ulp.h>


#define SWITCH_RED_PT 125
#define SWITCH_RED_RATE 90000
#define SWITCH_ULPFEC_PT 126
#define SWITCH_ULPFEC_RATE 90000

#define SWITCH_RED_FEC_BLOCKS_TO_SEND	10
#define SWITCH_RED_FEC_RTT_THRESHOLD_MS	50  /*TX*/
#define SWITCH_RED_FEC_RTT_THRESHOLD_MS_RX 100 /*RX*/
#define SWITCH_RED_FEC_TX_MAX_OVERHEAD 255 /* 100% */
#define SWITCH_RED_FEC_RE_RECOVER_SEQ_SHIFT	20
#define SWITCH_RED_FEC_RE_RECOVER_RECURSIVE_CALL_N	2
#define SWITCH_RTP_FRAME_FEC_MAX 50 /* TX */
#define SWITCH_RTP_FEC_COVER 2 /* TX, min gk */

typedef struct switch_red_s {
	switch_rtp_packet_t	*redundancy;
	uint16_t			redundancy_len;
	struct switch_red_s *next;				/* points to next entry in hash with same seq (snb) */
	uint8_t				n;					/* number of entries following this entry in hash in the bucket containing this entry including this entry, 1 if there is only this single entry or this is last entry */
	struct switch_red_s *up;				/* points to next entry sorted by increasing sequence */
} switch_red_t;

/**
 * Generate RTP header for RED.
 * Return pointer to RED packet or NULL on failure.
 * Timestamp, ssrc and seq are copied from @rtp header.
 * If @inline_where is NULL, then allocate memory and return pointer to it, otherwise
 * generate data in the place pointed to by @inline_where (must be >= 12 octets in length). 
 */
SWITCH_DECLARE(switch_rtp_hdr_t*) switch_red_gen_red_hdr(switch_rtp_packet_t *rtp, uint8_t red_pt, void *inline_where);

/**
 * Return the length of RED payload (encodings + their payloads), i.e length of RED packet less header.
 * @rtp_len - length of the RTP payload (RTP packet excluding header)
 * @redundancy_len - total length of the Redundant Encoding, in case of ULPFEC this is total FEC packet length (including FEC Header)
 */
SWITCH_DECLARE(uint16_t) switch_red_data_len(uint16_t rtp_len, uint16_t redundany_len);

/**
 * Generate RED packet payload: Primary Encoding Block Header and payload (media), plus optional Redundant Encoding Block Header and payload (FEC).
 *
 * Return pointer to RED payload or NULL on failure.
 * If @inline_where is NULL, then allocate memory and return pointer to it, otherwise
 * generate data in the place pointed to by @inline_where (must be >= ulp_red_data_len() in length).
 *
 * @rtp - points to media (excluding RTP header)
 * @rtp_len - length of the RTP payload (RTP packet excluding header)
 * @primary_pt - payload type of Primary Encoding, should be same as the payload of the transmitted RTP media packet
 * @redundancy - points to redundant data, in case of ULPFEC scheme it is FEC packet (total packet including FEC Header + all levels (header + payload))
 * @redundancy_len - total length of the redundancy (in case of FEC it is length of the packet including FEC header and all protection levels [header + payload])
 * @redundancy_pt - payload of the Redundant Encoding, for ULPFEC it is 127
 * @red_ts_off_diff - TS to be copied into Redundant Encoding ts field
 * @total_data_len - (out) on successful RED payload generation it is set to total length of the RED payload (excluding RED RTP header),
 *				on error it is set to 0 (though this is NULL returned value that should be checked)
 * @inline_where - if NULL, the memory needed for RED packet payload (excluding RED RTP header) is allocated, otherwise payload is created in memory
 *				pointed to by that (so it must be at least ulp_red_data_len()) 
 */
SWITCH_DECLARE(void*) switch_red_gen_red_data(switch_rtp_t *rtp_session, void *rtp, uint16_t rtp_len, uint8_t primary_pt, void *redundancy, uint16_t redundancy_len, uint8_t redundancy_pt, uint16_t red_ts_off_diff, uint16_t *total_data_len, void *inline_where);

/**
 * Generate RED packet: RED header + data (encodings + their payloads).
 *
 * Return pointer to RED packet or NULL on failure.
 *
 * @rtp	- points to media (including RTP header if @clean_hdr is not 0, excluding RTP header otherwise)
 * @rtp_len	- length of the RTP payload (RTP packet __excluding__ header)
 * @rtp_pt - PT of the RTP (to be put into Primary Encoding Block Header)
 * @clean_hdr - if 0, RED header is created from @rtp (which should point to RTP packet with header), otherwise memset RED header to 0
 * @redundancy - points to redundant data, in case of ULPFEC scheme it is FEC packet (total packet including FEC Header + all levels (header + payload))
 * @redundancy_len - total length of the redundancy (in case of FEC it is length of the packet including FEC header and all protection levels [header + payload])
 * @redundancy_pt - payload of the Redundant Encoding, for ULPFEC it is 127
 * @red_ts_off_diff - TS to be copied into Redundant Encoding ts field
 * @total_len - (out) on successful RED generation is set to total length of the RED packet (including RTP header),
 *				on error is set to 0 (though this is NULL returned value what indicates failure)
 * @red_pt - PT field in RTP header of created RED packet
 * @inline_where - if NULL, all needed memory for RED packet is allocated, otherwise packet is created in memory
 *				pointed to by that (so it must be at least ulp_red_data_len() + sizeof(RTP header)  
 */
SWITCH_DECLARE(switch_rtp_packet_t*) switch_red_gen_pkt(switch_rtp_t *rtp_session, void *rtp, uint16_t rtp_len, uint8_t rtp_pt, uint8_t clean_hdr, void *redundancy, uint16_t redundancy_len, uint8_t redundancy_pt, uint16_t red_ts_off_diff, uint16_t *total_len, uint8_t red_pt, void *inline_where);

/**
 * Generate RED packet with clean header: zeroed RED header + data (encodings + their payloads).
 *
 * Return pointer to RED packet or NULL on failure.
 *
 * @rtp	- points to media payload (media excluding RTP header)
 * @rtp_len	- length of the RTP payload (RTP packet __excluding__ header)
 * @rtp_pt - PT of the RTP (to be put into Primary Encoding Block Header)
 * @redundancy - points to redundant data, in case of ULPFEC scheme it is FEC packet (total packet including FEC Header + all levels (header + payload))
 * @redundancy_len - total length of the redundancy (in case of FEC it is length of the packet including FEC header and all protection levels [header + payload])
 * @redundancy_pt - payload of the Redundant Encoding, for ULPFEC it is 127
 * @red_ts_off_diff - TS to be copied into Redundant Encoding ts field
 * @total_len - (out) on successful RED generation is set to total length of the RED packet (including RTP header),
 *				on error is set to 0 (though this is NULL returned value what indicates failure)
 * @red_pt - PT field in RTP header of created RED packet
 * @inline_where - if NULL, all needed memory for RED packet is allocated, otherwise packet is created in memory
 *				pointed to by that (so it must be at least ulp_red_data_len() + sizeof(RTP header)  
 */
SWITCH_DECLARE(switch_rtp_packet_t*) switch_red_gen_pkt_clean_header(switch_rtp_t *rtp_session, void *rtp, uint16_t rtp_len, uint8_t rtp_pt, void *redundancy, uint16_t redundancy_len, uint8_t redundancy_pt, uint16_t red_ts_off_diff, uint16_t *total_len, uint8_t red_pt, void *inline_where);

/**
 * Generate RED packet with populated header: RED header + data (encodings + their payloads).
 *
 * Return pointer to RED packet or NULL on failure.
 *
 * @rtp	- points to RTP media (including RTP header)
 * @rtp_len	- length of the RTP payload (RTP packet __excluding__ header)
 * @rtp_pt - PT of the RTP (to be put into Primary Encoding Block Header)
 * @redundancy - points to redundant data, in case of ULPFEC scheme it is FEC packet (total packet including FEC Header + all levels (header + payload))
 * @redundancy_len - total length of the redundancy (in case of FEC it is length of the packet including FEC header and all protection levels [header + payload])
 * @redundancy_pt - payload of the Redundant Encoding, for ULPFEC it is 127
 * @red_ts_off_diff - TS to be copied into Redundant Encoding ts field
 * @total_len - (out) on successful RED generation is set to total length of the RED packet (including RTP header),
 *				on error is set to 0 (though this is NULL returned value what indicates failure)
 * @red_pt - PT field in RTP header of created RED packet
 * @inline_where - if NULL, all needed memory for RED packet is allocated, otherwise packet is created in memory
 *				pointed to by that (so it must be at least ulp_red_data_len() + sizeof(RTP header)  
 */
SWITCH_DECLARE(switch_rtp_packet_t*) switch_red_gen_pkt_with_header(switch_rtp_t *rtp_session, void *rtp, uint16_t rtp_len, uint8_t rtp_pt, void *redundancy, uint16_t redundancy_len, uint8_t redundancy_pt, uint16_t red_ts_off_diff, uint16_t *total_len, uint8_t red_pt, void *inline_where);

/**
 * Depacketize RED packet into Primary and Redundant encodings.
 *
 * Return pointer to the Primary Encoding (media), on failure return NULL.
 *
 * Note:
 *
 * RFC 5109: 10.3
 * In the payload format for redundant coding as specified by RFC 2198,
 * the marker bit is lost as soon as the primary coding is carried in the RED packets.
 * So the marker bit cannot be recovered regardless of whether or not the FEC is used.
 *
 * RFC 2198: 4  Limitations
 * The RTP marker bit is not preserved for redundant data blocks. Hence if the primary
 * (containing this marker) is lost, the marker is lost.
 *
 * @rtp - (in) should point to RED packet (including header),
 *			on return @rtp is set to unpacked Primary Encoding. If @inline_where is NULL, then Primary Encoding
 *			is copied to new malloced memory, otherwise Primary Encoding is copied to memory pointed to by @inline_where
 * @rtp_len - (in/out) on the call to this method @rtp_len should be set to length of the RED payload (excluding header),
 *				on return @rtp_len is set to the length of the extracted Primary Encoding (excluding header)
 * @redundancy - (in/out) on the return points to the extracted Redundant Encoding if any. Check @redundancy_len
 *					to know if there was Redundant Encoding
 * @redundancy_len - (out) on the return points to the total length of the Redundant Encoding (retrieved from Redundant Encoding Block Header),
 *						zero if there was no Redundant Encoding
 * @red_ts_off_diff - (out) timestamp offset field from Redundant Encoding Block Header (redundant encoding's timestamp offset from Primary Encoding),
 *						zero if there was no Redundant Encoding
 * @inline_where - configures where to extract Primary Encoding. If @inline_where is NULL, then Primary Encoding
 *					is copied to the new malloced memory, otherwise Primary Encoding is copied to the memory pointed to by @inline_where
 *					(so it can't overlap with @rtp and must be long enough, if it is not shorter than original RED packet, then this will work Okay)
 * @ulpfec_pt - PT of Redundant Encoding - must be known, becuse Chromium sends both media and FEC as Primary Encoding (MSB bit not set) but uses 4/1 bytes for redundant header (4 bytes for FEC even if first byte of encoding header says it is Primary) 
 */
SWITCH_DECLARE(switch_rtp_packet_t*) switch_red_depacketize(switch_core_session_t* session, ulp_policy_t *ulp_policy, switch_rtp_hdr_t *rtp, uint16_t *rtp_len, void **redundancy, uint16_t *redundancy_len, uint16_t *red_ts_off_diff, void* inline_where, uint8_t ulpfec_pt);

/**
 * Recovery procedure. Recover RTP packet with sequence @seq.
 * Return 1 if successfully recovered RTP packet or 0 if recovery was not possible.
 * If only partial recovery was possible @partial is set to 1 on the return,
 * otherwise @*partial is set to 0 indicating complete recovery.
 * @ssrc - host ordered SSRC of the to-be-recovered-packet
 */
SWITCH_DECLARE(uint8_t) switch_rtp_recover(ulp_policy_t *ulp_policy, uint16_t seq, ulp_status_t *ulp_err, switch_rtp_packet_t *recovered_packet, uint32_t ssrc);

/**
 * Parse ULPFEC packet and update @ulp_policy.
 */
SWITCH_DECLARE(switch_status_t) switch_red_ulpfec_parse(ulp_policy_t *ulp_policy, void *redundancy, uint16_t redundancy_len);

#endif // SWITCH_RED_H
